#include <stdio.h>
#include <signal.h>
#include <errno.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netdb.h>


#define PORT            5555
#define MESSAGE         "Yow!!! Hasta la vista, baby!"
#define SERVERHOST      "localhost"
#define MAXMSG  512

volatile int runserver;

volatile sig_atomic_t fatal_error_in_progress = 0;	// azert sig_atomic_t, hogy iras es olvasas kozben ne legyen gaz...

volatile pid_t pid_server;

volatile int sock;
volatile int cl_sock;



void sigh_client(int sig) {
    /* Since this handler is established for more than one kind of signal,
       it might still get invoked recursively by delivery of some other kind
       of signal.  Use a static variable to keep track of that. */
    if (fatal_error_in_progress)
	raise(sig);
    fatal_error_in_progress = 1;



    /* Now do the clean up actions:
       - reset terminal modes
       - kill child processes
       - remove lock files */

    fprintf(stderr, "client: caught signal\n");
    
    runserver=0;

    kill(pid_server,SIGQUIT);

    fprintf(stderr, "client: closing socket\n");

    if(shutdown(cl_sock,2)) {
	perror("close socket");
    } else {
	fprintf(stderr, "client: done\n");
    };

    /* Now reraise the signal.  We reactivate the signal's
       default handling, which is to terminate the process.
       We could just call `exit' or `abort',
       but reraising the signal sets the return status
       from the process correctly. */
    signal(sig, SIG_DFL);
    raise(sig);
}

volatile sig_atomic_t sigh_server_in_progress = 0;	// azert sig_atomic_t, hogy iras es olvasas kozben ne legyen gaz...

void sigh_server(int sig) {
    /* Since this handler is established for more than one kind of signal,
       it might still get invoked recursively by delivery of some other kind
       of signal.  Use a static variable to keep track of that. */
    if (sigh_server_in_progress)
	raise(sig);
    sigh_server_in_progress = 1;



    /* Now do the clean up actions:
       - reset terminal modes
       - kill child processes
       - remove lock files */

    fprintf(stderr, "server: caught signal\n");
    
    runserver=0;

    fprintf(stderr, "server: closing socket\n");

    if(shutdown(sock,2)) {
	perror("close socket");
    } else {
	fprintf(stderr, "server: done\n");
    };

    fprintf(stderr, "server: done\n");

    /* Now reraise the signal.  We reactivate the signal's
       default handling, which is to terminate the process.
       We could just call `exit' or `abort',
       but reraising the signal sets the return status
       from the process correctly. */
    signal(sig, SIG_DFL);
    raise(sig);
}


int make_socket(uint16_t port) {
    int sock;
    int opt=1;
    struct sockaddr_in name;

    /* Create the socket. */
    sock = socket(PF_INET, SOCK_STREAM, 0);
    if (sock < 0) {
	perror("socket");
	exit(EXIT_FAILURE);
    }

    if(setsockopt(sock,SOL_SOCKET,SO_KEEPALIVE,&opt,sizeof(int))) {
	perror("setsockopt");
	exit(EXIT_FAILURE);
    };

    if(setsockopt(sock,SOL_SOCKET,SO_REUSEADDR,&opt,sizeof(int))) {
	perror("setsockopt");
	exit(EXIT_FAILURE);
    };

    /* Give the socket a name. */
    name.sin_family = AF_INET;
    name.sin_port = htons(port);
    name.sin_addr.s_addr = htonl(INADDR_ANY);
    name.sin_port = htons(port);
    name.sin_addr.s_addr = htonl(INADDR_ANY);
    if (bind(sock, (struct sockaddr *) &name, sizeof(name)) < 0) {
	perror("bind");
	exit(EXIT_FAILURE);
    }

    return sock;
}

void
init_sockaddr(struct sockaddr_in *name,
	      const char *hostname, uint16_t port) 
{
    struct hostent *hostinfo;

    name->sin_family = AF_INET;
    name->sin_port = htons(port);
    hostinfo = gethostbyname(hostname);
    if (hostinfo == NULL)
	if (hostinfo == NULL) {
	    fprintf(stderr, "Unknown host %s.\n", hostname);
	    exit(EXIT_FAILURE);
	}
    name->sin_addr = *(struct in_addr *) hostinfo->h_addr;
}

void write_to_server(int filedes)
{
    int nbytes;

    nbytes = write(filedes, MESSAGE, strlen(MESSAGE) + 1);
    if (nbytes < 0) {
	perror("write");
	exit(EXIT_FAILURE);
    }
}

int read_from_client(int filedes)
{
    char buffer[MAXMSG];
    int nbytes;

    nbytes = read(filedes, buffer, MAXMSG);
    if (nbytes < 0) {
	/* Read error. */
	perror("read");
//      exit(EXIT_FAILURE);
    } else if (nbytes == 0)
	/* End-of-file. */
	return -1;
    else {
	/* Data read. */
	fprintf(stderr, "Server: got message: `%s'\n", buffer);
	return 0;
    }
}

int server(void)
{
    extern int make_socket(uint16_t port);
    fd_set active_fd_set, read_fd_set;
    int i;
    struct sockaddr_in clientname;
    size_t size;

    signal(SIGTERM, sigh_server);
    signal(SIGINT, sigh_server);
    signal(SIGQUIT, sigh_server);
    signal(SIGHUP, sigh_server);
    signal(SIGKILL, sigh_server);

    /* Create the socket and set it up to accept connections. */
    sock = make_socket(PORT);
    if (listen(sock, 1) < 0) {
	perror("listen");
	exit(EXIT_FAILURE);
    }

    /* Initialize the set of active sockets. */
    FD_ZERO(&active_fd_set);
    FD_SET(sock, &active_fd_set);

    runserver=1;

    while (runserver) {
	/* Block until input arrives on one or more active sockets. */
	read_fd_set = active_fd_set;
	if (select(FD_SETSIZE, &read_fd_set, NULL, NULL, NULL) < 0) {
	    perror("select");
	    exit(EXIT_FAILURE);
	}
	/* Service all the sockets with input pending. */
	for (i = 0; i < FD_SETSIZE; ++i)
	    if (FD_ISSET(i, &read_fd_set)) {
		if (i == sock) {
		    /* Connection request on original socket. */
		    int new;
		    size = sizeof(clientname);
		    new = accept(sock,
				 (struct sockaddr *) &clientname, &size);
		    if (new < 0) {
			perror("accept");
			exit(EXIT_FAILURE);
		    }
		    FD_SET(new, &active_fd_set);
		    fprintf(stderr,
			    "Server: connect from host %s, port %hd.\n",
			    inet_ntoa(clientname.sin_addr),
			    ntohs(clientname.sin_port));
		} else {
		    /* Data arriving on an already-connected socket. */
		    if (read_from_client(i) < 0) {
			fprintf(stderr, "Server: connection lost\n");
			close(i);
			FD_CLR(i, &active_fd_set);
		    } else {	// meg nyitva van a socket

			write_to_server(i);
		    };

		}
	    }
    }
    
    close(sock); // socket kigyilkolasa
    
}

int main(void)
{
    extern void init_sockaddr(struct sockaddr_in *name,
			      const char *hostname, uint16_t port);

    char i;
    struct sockaddr_in servername;

    pid_server=fork();

    if(pid_server==-1) {
	perror("fork");
	exit(EXIT_FAILURE);
    };

    if (pid_server==0) { // mi vagyunk a szerver
	fprintf(stderr,"attempting to create server...\n");
	server();
	fprintf(stderr,"server exiting\n");
	exit(EXIT_SUCCESS);
    };

    // itt else kell majd. mi vagyunk a kliens

    signal(SIGTERM, sigh_client);
    signal(SIGINT, sigh_client);
    signal(SIGQUIT, sigh_client);
    signal(SIGHUP, sigh_client);
    signal(SIGKILL, sigh_client);


    sleep(1);

    /* Create the socket. */
    cl_sock = socket(PF_INET, SOCK_STREAM, 0);
    if (cl_sock < 0) {
	perror("socket (client)");
	exit(EXIT_FAILURE);
    }

    /* Connect to the server. */
    init_sockaddr(&servername, SERVERHOST, PORT);
    if (0 > connect(cl_sock,
		    (struct sockaddr *) &servername, sizeof(servername))) {
	perror("connect (client)");
	exit(EXIT_FAILURE);
    }

    /* Send data to the server. */
    while (1) {

	write_to_server(cl_sock);


	if (read_from_client(cl_sock) < 0) {
	    fprintf(stderr,"client: server disconnected\n");
	    close(cl_sock);
	    exit(EXIT_FAILURE);
	    
//      FD_CLR(i, &active_fd_set);
	}

	getc(stdin);

    };

    close(cl_sock);
    exit(EXIT_SUCCESS);
}
