//
// KUKAC jatek - szglab5 halozatok feladat
//
// (C)opyright by McRee [mcree@freemail.hu]
//

#include "globals.h"
#include "inetfunct.h"
#include <sys/time.h>
#include <stdio.h>
#include <signal.h>
#include <errno.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netdb.h>

volatile sig_atomic_t sigh_server_in_progress = 0;	// azert sig_atomic_t, hogy iras es olvasas kozben ne legyen gaz...
volatile int runserver;
volatile int sv_sock;

t_level level;

void killplayer(int player)
{
    int u, i;
    for (u = 0; u < MAX_Y; u++) {
	for (i = 0; i < MAX_X; i++) {
	    if (level[i][u] == player)
		level[i][u] = 0;
	};
    };
};

void sigh_server(int sig)
{
    /* Since this handler is established for more than one kind of signal,
       it might still get invoked recursively by delivery of some other kind
       of signal.  Use a static variable to keep track of that. */
    if (sigh_server_in_progress)
	raise(sig);
    sigh_server_in_progress = 1;

    /* Now do the clean up actions:
       - reset terminal modes
       - kill child processes
       - remove lock files */

    fprintf(stderr, "server: caught signal\n");

    runserver = 0;

    fprintf(stderr, "server: closing socket\n");

    if (shutdown(sv_sock, 2)) {
	perror("close socket");
    } else {
	fprintf(stderr, "server: done\n");
    };

    /* Now reraise the signal.  We reactivate the signal's
       default handling, which is to terminate the process.
       We could just call `exit' or `abort',
       but reraising the signal sets the return status
       from the process correctly. */
    signal(sig, SIG_DFL);
    raise(sig);
}

t_players players[MAXPLAYERS];

void dropplayer(int player)
{
    int i;			// dummy
    int found = -1;		// player found

    // serach if player is already registered
    for (i = 0; i < MAXPLAYERS; i++) {
	if (players[i].player == player) {
	    found = i;
	};
    };

    if (found < 0) {
	perror("server: trying to drop unregistered player");
	exit(EXIT_FAILURE);
    } else {
	fprintf(stderr, "server: dropping player (%i)\n", player);
	killplayer(found);
	players[found].player = 0;
    };

};

int getplayer(int player)
{
    int i;			// dummy
    int found = -1;		// player found

    // serach if player is already registered
    for (i = 0; i < MAXPLAYERS; i++) {
	if (players[i].player == player) {
	    found = i;
	};
    };

    if (found < 0) {
	perror("server: trying to get unregistered player");
	exit(EXIT_FAILURE);
    } else {
	return found;
    };

};

void updateplayer(int player, int x, int y, int dir, int action, int lives, int score)
{
    int i;			// dummy
    int found = -1;		// player found

    // serach if player is already registered
    for (i = 0; i < MAXPLAYERS; i++) {
	if (players[i].player == player) {
	    found = i;
	};
    };

    if (found < 0) {
	// serach for unused player slot
	for (i = 0; i < MAXPLAYERS; i++) {
	    if (players[i].player == 0) {
		found = i;
	    };
	};

	if (found < 0) {
	    perror("server: no unused player slots left. sorry.");
	    exit(EXIT_FAILURE);
	};
    };

    // update player entry
    players[found].player = player;
    if (x != -1)
	players[found].x = x;
    if (y != -1)
	players[found].y = y;
    if (dir != -1)
	players[found].dir = dir;
    if (action != -1)
	players[found].action = action;
    if (lives != -1)
	players[found].lives = lives;
    if (score != -1)
	players[found].score = score;

};

int read_from_client(int filedes)
{
    char buffer[MAXMSG];
    int nbytes, i, dir;

    nbytes = read(filedes, buffer, MAXMSG);
    if (nbytes < 0) {
	/* Read error. */
	perror("server: read from client");
//      exit(EXIT_FAILURE);
	return -1;
    } else if (nbytes == 0)
	/* End-of-file. */
	return -1;
    else {
	/* Data read. */
//      fprintf(stderr, "server: got message (from %x): `%s'\n", filedes, buffer);

	dir = atoi(buffer);
	if (players[getplayer(filedes)].dir + dir != 0) {	// if opposide direction is not choosen
	    updateplayer(filedes, -1, -1, dir, -1, -1, -1);
	};
	return 0;
    }
}

//#define MESSAGE "sv"
int write_to_client(int filedes)
{
    int nbytes, i;

    nbytes = write(filedes, players, sizeof(t_players) * MAXPLAYERS + 1);
//    fprintf(stderr, "server: send: message size: %i to client %i\n", sizeof(t_players) * MAXPLAYERS + 1, filedes);
    if (nbytes < 0) {
	perror("write to client");
//      exit(EXIT_FAILURE);
    }

/*    for (i = 0; i < MAXPLAYERS; i++) {
	if (players[i].disconn == 1) {
	    dropplayer(i);
	};
    };
*/
    return nbytes;
}


void server(void)
{
    fd_set active_fd_set, read_fd_set, write_fd_set;	// file descriptor sets
    int i, n, u, uu, x, y;	// dummy
//    time_t last,now;
    struct sockaddr_in clientname;	// tmp variable used for initializing a new client onnection
    size_t size;		// dummy
    struct timespec servertick, remaining;	// size one server tick
    int nbytes;


    fprintf(stderr, "server: starting\n");

    memset(&players, 0, sizeof(t_players) * MAXPLAYERS);	// clear players table
    memset(&level, 0, sizeof(t_level));	// clear table

    // setup default server delay
    servertick.tv_sec = 0;
    servertick.tv_nsec = 200000;	// nsec must be <1000000000

    // setup signal handling
    signal(SIGTERM, sigh_server);
    signal(SIGINT, sigh_server);
    signal(SIGQUIT, sigh_server);
    signal(SIGHUP, sigh_server);
    signal(SIGKILL, sigh_server);

    fprintf(stderr, "server: binding to port %i\n", PORT);

    // init and start listening on socket
    sv_sock = make_socket(PORT);	// create socket and bind to specified port
    if (listen(sv_sock, 1) < 0) {	// listen to connections (accept 1 connection pending)
	perror("server: listen");
	exit(EXIT_FAILURE);
    }

    fprintf(stderr, "server: listening\n", PORT);

    // Initialize the set of active sockets
    FD_ZERO(&active_fd_set);	// empty set
    FD_SET(sv_sock, &active_fd_set);	// insert our socket to the set (we accept connections on it)

    runserver = 1;		// the server is considered running

    while (runserver) {		// 'loop forever'
	// wait for I/O, read client input, periodically update (write to) clients
	read_fd_set = active_fd_set;
	write_fd_set = active_fd_set;
	FD_CLR(sv_sock, &write_fd_set);	// remove server socket from writeable set

	nanosleep(&servertick, &remaining);

	if (select(FD_SETSIZE, &read_fd_set, &write_fd_set, NULL, 0) < 0) {
	    perror("server: select");
	    exit(EXIT_FAILURE);
	}
	// Service all the sockets with input pending.
	for (i = 0; i < FD_SETSIZE; ++i)
	    if (FD_ISSET(i, &read_fd_set)) {
		if (i == sv_sock) {
		    // Connection request on original socket.
		    int new;
		    size = sizeof(clientname);
		    new = accept(sv_sock, (struct sockaddr *) &clientname, &size);
		    if (new < 0) {
			perror("server: accept connection");
			exit(EXIT_FAILURE);
		    }
		    FD_SET(new, &active_fd_set);
		    fprintf(stderr,
			    "server: client (%i) connect from host %s, port %hd.\n",
			    new, inet_ntoa(clientname.sin_addr), ntohs(clientname.sin_port));

		    x = rand() % MAX_X;
		    y = rand() % MAX_Y;

		    updateplayer(new, x, y, rand() % 4, ACT_INV, 0, 0);
		    players[getplayer(new)].counter = INV_LENGTH;

//                  level[x][y] = getplayer(new);

		    nbytes = write(new, level, sizeof(t_level) + 1);
		    fprintf(stderr, "server: send: message size: %i to client %i\n", sizeof(t_players) * MAXPLAYERS + 1,
			    new);
//    nbytes = write(filedes, MESSAGE, strlen(MESSAGE) + 1);
		    if (nbytes < 0) {
			perror("write");
			exit(EXIT_FAILURE);
		    }

		} else {
		    // Data arriving on an already-connected socket.
		    if (read_from_client(i) < 0) {
			fprintf(stderr, "server: client (%i) connection lost\n", i);
			players[getplayer(i)].disconn = 2;
//                      killplayer(getplayer(i));
			dropplayer(i);
			FD_CLR(i, &active_fd_set);
			close(i);
		    };

		}
	    }
//      if (level[0][0] != 0)
//          exit(-1);


	// Service all the sockets with output pending. 
	if ((n++) > 5) {

	    // run trough players
	    //
	    // STEPS
	    //
	    for (u = 0; u < MAXPLAYERS; u++) {
		if (players[u].player != 0) {

		    // change player status
		    if (players[u].action == ACT_DIE) {
			players[u].action = ACT_INV;
			players[u].counter = INV_LENGTH;
		    } else if (players[u].action == ACT_INV) {
			players[u].counter--;
			if (players[u].counter < 0) {
			    players[u].action = ACT_RESPAWN;
			};
		    } else {
			players[u].action = ACT_GO;
		    };

		    // move to direction
		    x = players[u].x;
		    y = players[u].y;
		    if (players[u].player != 0)
			switch (players[u].dir) {
			case DIR_LEFT:
			    x--;
			    break;
			case DIR_RIGHT:
			    x++;
			    break;
			case DIR_UP:
			    y--;
			    break;
			case DIR_DOWN:
			    y++;
			    break;
			};
		    if (x == MAX_X)
			x = 0;
		    if (x < 0)
			x = MAX_X - 1;
		    if (y == MAX_Y)
			y = 0;
		    if (y < 0)
			y = MAX_Y - 1;

		    players[u].x = x;
		    players[u].y = y;

//                  fprintf(stderr,
//                          "server: registered player %i (slot %i): x=%i y=%i dir=%i action=%i lives=%i score=%i\n",
//                          players[u].player, u, players[u].x, players[u].y, players[u].dir, players[u].action,
//                          players[u].lives, players[u].score);
		};
	    };

	    // test for HEAD-HEAD collisions
	    for (u = 0; u < MAXPLAYERS; u++) {
		if (players[u].player != 0) {
		    for (uu = 0; uu < MAXPLAYERS; uu++) {
			if ((players[u].action == ACT_GO) && ((players[u].x == players[uu].x)
							      && (players[u].y == players[uu].y) && (u != uu)
							      && (players[uu].player != 0)
							      || (level[players[u].x][players[u].y] == uu)
							      && (level[players[uu].x][players[uu].y] == u && u != uu)
			    )) {
			    players[u].action = ACT_DIE;
			    players[u].lives--;	// self descruction penalty
			    players[u].score -= 10;	// self descruction penalty
			};
		    };
		};
	    };

	    // test for HEAD->BODY collisions
	    for (u = 0; u < MAXPLAYERS; u++) {
		if (players[u].player != 0) {
		    if ((level[players[u].x][players[u].y] != 0) && (players[u].action == ACT_GO)) {
			players[u].action = ACT_DIE;
			if (level[players[u].x][players[u].y] != u) {	// collide ourself
			    players[level[players[u].x][players[u].y]].lives++;	// grant enemy a frag
			    players[level[players[u].x][players[u].y]].score += 100;
			} else {
			    players[u].lives--;	// self descruction penalty
			    players[u].score -= 100;
			};
			//killplayer(u);
		    } else {
			level[players[u].x][players[u].y] = u;
			players[u].score++;
		    };
		};
	    };



	    // kill dead players
	    for (u = 0; u < MAXPLAYERS; u++) {
		if (players[u].action == ACT_DIE)
		    killplayer(u);
	    };

	    // timeout disconnected players
	    for (u = 0; u < MAXPLAYERS; u++) {
		if (players[u].disconn > 0)
		    players[u].disconn--;
	    };

	    n = 0;
	    for (i = 0; i < FD_SETSIZE; ++i)
		if (FD_ISSET(i, &write_fd_set)) {
		    if (write_to_client(i) < 0) {
//                      fprintf(stderr, "server: client (%i) connection lost\n", i);
//                      players[getplayer(i)].disconn = 2;
//                      killplayer(getplayer(i));
//                      dropplayer(i);
//                      FD_CLR(i, &active_fd_set);
//                      close(i);
		    };
		}
	};
    }

    close(sv_sock);		// kill socket

}
